/**
 * Functionality for the running of the Cydia Pomonella model. 
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses />.
 * 
 * Created 2024-10-30
 * @author Tor-Einar Skog <tor-einar.skog@nibio.no>
 */

class CydiaPomonella 
{
    weatherStationDataURL = undefined;
    gridDataURL = undefined;
    TIMEZONE_OSLO = "Europe/Oslo";
    DATE_FORMAT = "YYYY-MM-DD";
    HEATSUM_BASE_TEMP = 10.0;
    WEATHER_DATASOURCE_TYPE_COORDINATES = "coordinates";
    WEATHER_DATASOURCE_TYPE_WEATHERSTATION = "weatherstation";
    progressIconRow = document.getElementById("progressIconRow");
    

    appleWeatherStations = [
        { "id": 11, "name": "Apelsvoll" },
        { "id": 12, "name": "Balestrand" },
        { "id": 13, "name": "Bø" },
        { "id": 86, "name": "Darbu" },
        { "id": 19, "name": "Gjerpen" },
        { "id": 21, "name": "Gvarv" },
        { "id": 22, "name": "Hjelmeland" },
        { "id": 25, "name": "Hønefoss" },
        { "id": 27, "name": "Kise" },
        { "id": 29, "name": "Landvik" },
        { "id": 30, "name": "Lier" },
        { "id": 65, "name": "Ljøsne" },
        { "id": 32, "name": "Lyngdal" },
        { "id": 35, "name": "Njøs" },
        { "id": 41, "name": "Rygge" },
        { "id": 42, "name": "Sande" },
        { "id": 131, "name": "Sandefjord" },
        { "id": 64, "name": "Slinde" },
        { "id": 47, "name": "Svelvik" },
        { "id": 91, "name": "Søve" },
        { "id": 61, "name": "Åsbakken" }
    ];

    appleWeatherStationIds = this.appleWeatherStations.map(ws=>ws.id);

    constructor(lmtServicesURL, gridDataURL){
        this.weatherStationDataURL = lmtServicesURL + "/rest/vips/getdata/forecastfallback";
        this.gridDataURL = gridDataURL;
    }

    renderWeatherstations = function () {
        let wsSelect = document.getElementById("weatherStationId");
        this.appleWeatherStations.forEach((ws) => {
            wsSelect.options[wsSelect.options.length] = new Option(ws["name"], ws["id"]);
        });
    }

    getSelectedWeatherstation()
    {
        let wsSelect = document.getElementById("weatherStationId");
        return wsSelect.options[wsSelect.selectedIndex].value != "-1" ? wsSelect.options[wsSelect.selectedIndex].value : undefined;
    }

    getSelectedCoordinate()
    {
        let longitude = parseFloat(document.getElementById("longitude").value);
        let latitude = parseFloat(document.getElementById("latitude").value);
        return (Number.isNaN(longitude) || Number.isNaN(latitude) ? undefined : [longitude, latitude]);
    }

    getSelectedWeatherdataSourceType()
    {
        return document.getElementById("coordinates").checked ? this.WEATHER_DATASOURCE_TYPE_COORDINATES : document.getElementById("weatherstation").checked ? this.WEATHER_DATASOURCE_TYPE_WEATHERSTATION : undefined;
    }

    async runModel() {
        this.progressIconRow.style.display = "block";
        let selectedWeatherStationId = this.getSelectedWeatherstation();
        let selectedCoordinate = this.getSelectedCoordinate();
        if(selectedWeatherStationId == undefined && selectedCoordinate == undefined)
        {
            return;
        }

        let startDateHourly = this.getStartDateHourly();
        let endDateHourly = this.getEndDateHourly();
        try
        {
            let hourlyDataResponse = undefined;
            if(this.getSelectedWeatherdataSourceType() == this.WEATHER_DATASOURCE_TYPE_WEATHERSTATION && selectedWeatherStationId != undefined)
            {
                // Get the hourly data for the past X days
                hourlyDataResponse = await fetch(
                    this.weatherStationDataURL + "?weatherStationId=" + selectedWeatherStationId 
                    + "&elementMeasurementTypes[]=RR&elementMeasurementTypes[]=TM"
                    + "&timeZone="  + this.TIMEZONE_OSLO 
                    + "&startDate=" + startDateHourly.format(this.DATE_FORMAT) + "&startTime=0" 
                    + "&endDate=" + endDateHourly.format(this.DATE_FORMAT) + "&endTime=12"  
                    + "&logIntervalId=1"
                );
            }
            else if(this.getSelectedWeatherdataSourceType() == this.WEATHER_DATASOURCE_TYPE_COORDINATES && selectedCoordinate != undefined)
            {
                hourlyDataResponse = await fetch(
                    this.gridDataURL.replace("%s",selectedCoordinate[0] + "_" + selectedCoordinate[1]) //+ "?longitude=" + selectedCoordinate[0] + "&latitude=" + selectedCoordinate[1]
                    + "?elementMeasurementTypes[]=RR&elementMeasurementTypes[]=TM"
                    + "&timeZone="  + document.getElementById("timezone").value
                    + "&startDate=" + startDateHourly.format(this.DATE_FORMAT) + "&startTime=0" 
                    + "&endDate=" + endDateHourly.format(this.DATE_FORMAT) + "&endTime=12"  
                    + "&logIntervalId=1"
                );
            }
            else
            {
                console.error("Location not set. Aborting.")
                return;
            }
            if(! hourlyDataResponse || ! hourlyDataResponse.ok){
                throw new Error("Response status: " + hourlyDataResponse.status);
            }

            const hourlyData = await hourlyDataResponse.json();

            // Get the daily RR data for the past X days
            const dailyData = this.getDailyRRData(hourlyData);

            this.renderTable(hourlyData, dailyData);

            this.runHeatSumCalculations();
            
        }
        catch(error)
        {
            console.error(error.message);
        }
        this.progressIconRow.style.display = "none";
    }

    async runHeatSumCalculations()
    {
        document.getElementById("heatSumContainer").style.display="block";
        let selectedWeatherStationId = this.getSelectedWeatherstation();
        let selectedCoordinate = this.getSelectedCoordinate();
        if(selectedWeatherStationId == undefined && selectedCoordinate == undefined)
        {
            return;
        }
        // Get daily temperature data for the heat sum
        const heatSumStartDate = this.setToTZMidnight(moment(document.getElementById("heatSumStartDate").value));
        const heatSumEndDate = this.setToTZMidnight(moment().tz(this.TIMEZONE_OSLO).subtract(1,"days"));
        
        let heatSumDataResponse = undefined;
        if(this.getSelectedWeatherdataSourceType() == this.WEATHER_DATASOURCE_TYPE_WEATHERSTATION && selectedWeatherStationId != undefined)
        {
            heatSumDataResponse = await fetch(
                this.weatherStationDataURL + "?weatherStationId=" + selectedWeatherStationId 
                + "&elementMeasurementTypes[]=TM"
                + "&timeZone="  + this.TIMEZONE_OSLO 
                + "&startDate=" + heatSumStartDate.format(this.DATE_FORMAT) + "&startTime=0" 
                + "&endDate=" + heatSumEndDate.format(this.DATE_FORMAT) + "&endTime=23"  
                + "&logIntervalId=2"
            );
        }
        else if(this.getSelectedWeatherdataSourceType() == this.WEATHER_DATASOURCE_TYPE_COORDINATES && selectedCoordinate != undefined)
        {
            heatSumDataResponse = await fetch(
                this.gridDataURL.replace("%s",selectedCoordinate[0] + "_" + selectedCoordinate[1])// + "?longitude=" + selectedCoordinate[0] + "&latitude=" + selectedCoordinate[1]
                + "?elementMeasurementTypes[]=TM"
                + "&timeZone="  + document.getElementById("timezone").value
                + "&startDate=" + heatSumStartDate.format(this.DATE_FORMAT) + "&startTime=0" 
                + "&endDate=" + heatSumEndDate.format(this.DATE_FORMAT) + "&endTime=23"  
                + "&logIntervalId=2"
            );
        }
        else
        {
            console.error("Location not set. Aborting.")
            return;
        }

        if(! heatSumDataResponse.ok){
            throw new Error("Response status: " + heatSumDataResponse.status);
        }

        const heatSumData = await heatSumDataResponse.json();

        this.renderHeatSumResult(this.calculateHeatSum(heatSumData), heatSumData.length, heatSumStartDate, heatSumEndDate);
    }

    calculateHeatSum(heatSumData)
    {
        let heatSum = 0.0;
        heatSumData.forEach((obs) =>{
            heatSum += Math.max(0.0, (obs.value - this.HEATSUM_BASE_TEMP));
        });
        return heatSum;
    }

    renderHeatSumResult(heatSum, numberOfDays, heatSumStartDate, heatSumEndDate)
    {
        let weatherStationName = undefined;
        if(this.getSelectedWeatherdataSourceType() == this.WEATHER_DATASOURCE_TYPE_WEATHERSTATION)
        {
            const selectedWeatherstationId = parseInt(this.getSelectedWeatherstation());
            this.appleWeatherStations.forEach((ws)=>{
                if(ws.id == selectedWeatherstationId)
                {
                    weatherStationName =  ws.name;
                }
            });
        }

        const heatSumMsg = gettext("Heat sum, with a base temperature of") + " " + this.HEATSUM_BASE_TEMP + "&deg;C, " + gettext("calculated for") + " " 
        + (this.getSelectedWeatherdataSourceType() == this.WEATHER_DATASOURCE_TYPE_WEATHERSTATION ? weatherStationName + " " + gettext("weatherstation") : gettext("selected coordinates"))
        + " " + gettext("for the period starting with") + " " + heatSumStartDate.format(this.DATE_FORMAT) + " " + gettext("and ending with") + " " + heatSumEndDate.format(this.DATE_FORMAT)
        + " " + "(" + numberOfDays + " " + gettext("days") + ")"
        + " " + gettext("is") + ": <strong>" + heatSum.toFixed(1) + " " + gettext("day degrees") + "</strong>";
        
        document.getElementById("heatSumResult").innerHTML = heatSumMsg;
    }

    getDailyRRData(hourlyData)
    {
        let RRBuckets = {};
        for(let i=0; i< hourlyData.length;i++)
        {
            let obs = hourlyData[i];
            if(obs.elementMeasurementTypeId == "RR")
            {
                let date = this.setToTZMidnight(moment(obs.timeMeasured));
                if(RRBuckets[date.unix()] == null)
                {
                    RRBuckets[date.unix()] = [];
                }
                RRBuckets[date.unix()].push(obs.value);
            }
        }

        let dailyRRData = [];
        Object.keys(RRBuckets).forEach((key) => {
            let dateBucket = RRBuckets[key];
            let RRSum = dateBucket.reduce((accumulator, currentValue) => accumulator + currentValue, 0);
            let RR = {
                timeMeasured: moment.unix(key).toISOString(),
                elementMeasurementTypeId: "RR",
                logIntervalId: 1,
                value: RRSum
            }
            dailyRRData.push(RR);
        });
        
        return dailyRRData;
    }

    renderTable(hourlyData, dailyData)
    {
        document.getElementById("weatherData").style.display="table";
        let tableBody = document.getElementById("tableBody");
        tableBody.innerHTML = "";
        let currentDate = this.getStartDateHourly();
        let today = this.setToTZMidnight(moment());
        //console.info("today=" + today.format());
        while(currentDate < today)
        {
            //console.info("currentDate=" + currentDate.format());
            let nightTemps = this.getNightTempsForDate(moment(currentDate).tz(this.TIMEZONE_OSLO), hourlyData);
            //console.info(nightTemps);
            let rowHTML = "<tr><td>" + currentDate.format(this.DATE_FORMAT) + "</td>";
            
            for(let i=0; i<nightTemps.length;i++)
            {
                rowHTML += "<td>" + nightTemps[i].toFixed(1) + "</td>";
            }
            dailyData.forEach((obs) =>{
                if(moment(obs.timeMeasured).unix() == currentDate.unix())
                {
                    rowHTML += "<td>" + obs.value.toFixed(1) + "</td>";
                }
            });
            rowHTML += "</tr>"
            tableBody.innerHTML += rowHTML;
            currentDate.add(1,"days");
        }
    }

    getStartDateHourly = function()
    {
        let daysBack = undefined;
        document.getElementsByName("radioDays").forEach((radio)=>{
            if(radio.checked)
            {
                daysBack = radio.value;
            }
        });
        return this.setToTZMidnight(moment()).subtract(daysBack,"days");
    }

    getEndDateHourly()
    {
        return moment().tz(this.TIMEZONE_OSLO).hour(12);
    }

    /**
     * 
     * @param date 
     * @param {Array} weatherData 
     * @return array[temp@19,temp@20,temp@21,temp@22, temp@23]
     */
    getNightTempsForDate(date, weatherData)
    {
        // Create an array of the desired timestamps
        let timestamps = [];
        let nightTemps = [];
        for(let i=19;i<=23;i++)
        {
            timestamps.push(date.hour(i).minute(0).second(0).millisecond(0).unix());
        }
        weatherData.forEach((wd)=>{
            let timeMeasured = moment(wd.timeMeasured).unix();
            //console.info(timeMeasured);
            if(wd.elementMeasurementTypeId == "TM" && timestamps.indexOf(timeMeasured) >=0)
            {
                nightTemps[timestamps.indexOf(timeMeasured)] = wd.value;
                //console.info("Found " + wd.timeMeasured + "at index " + tsArray.indexOf(timeMeasured));
            }
        });
        return nightTemps;
        //console.info(tsArray);
    }

    setToTZMidnight(theDate)
    {
        return theDate.tz(this.TIMEZONE_OSLO).hour(0).minute(0).second(0).millisecond(0);
    }
}

export default CydiaPomonella;